(globalThis.TURBOPACK||(globalThis.TURBOPACK=[])).push(["object"==typeof document?document.currentScript:void 0,36557,c=>{c.v(t=>Promise.all(["static/chunks/057c9be8ac221ca0.js"].map(t=>c.l(t))).then(()=>t(7074)))}]);
